# -*- test-case-name: xmantissa.test.test_offering -*-
# Copyright 2008 Divmod, Inc. See LICENSE file for details
"""
Implemenation of Offerings, Mantissa's unit of installable application
functionality.
"""

from zope.interface import implements

from twisted import plugin
from twisted.python.components import registerAdapter

from nevow import inevow, loaders, rend, athena
from nevow.athena import expose

from epsilon.structlike import record

from axiom.store import Store
from axiom import item, userbase, attributes, substore
from axiom.dependency import installOn

from xmantissa import ixmantissa, plugins

class OfferingAlreadyInstalled(Exception):
    """
    Tried to install an offering, but an offering by the same name was
    already installed.

    This may mean someone tried to install the same offering twice, or
    that two unrelated offerings picked the same name and therefore
    conflict!  Oops.
    """

class Benefactor(object):
    """
    I implement a method of installing and removing chunks of
    functionality from a user's store.
    """

class Offering(record(
        'name description siteRequirements appPowerups installablePowerups '
        'loginInterfaces themes staticContentPath version',
        staticContentPath=None, version=None)):
    """
    A set of functionality which can be added to a Mantissa server.

    @see L{ixmantissa.IOffering}
    """
    implements(plugin.IPlugin, ixmantissa.IOffering)


class InstalledOffering(item.Item):
    typeName = 'mantissa_installed_offering'
    schemaVersion = 1

    offeringName = attributes.text(doc="""
    The name of the Offering to which this corresponds.
    """, allowNone=False)

    application = attributes.reference(doc="""
    A reference to the Application SubStore for this offering.
    """)

    def getOffering(self):
        """
        @return: the L{Offering} plugin object that corresponds to this object,
        or L{None}.
        """
        # XXX maybe we need to optimize this; it's called SUPER often, and this
        # is ghetto as hell, on the other hand, isn't that the plugin system's
        # fault?  dang, I don't know.
        for o in getOfferings():
            if o.name == self.offeringName:
                return o



def getOfferings():
    """
    Return the IOffering plugins available on this system.
    """
    return plugin.getPlugins(ixmantissa.IOffering, plugins)


class OfferingAdapter(object):
    """
    Implementation of L{ixmantissa.IOfferingTechnician} for
    L{axiom.store.Store}.

    @ivar _siteStore: The L{axiom.store.Store} being adapted.
    """
    implements(ixmantissa.IOfferingTechnician)

    def __init__(self, siteStore):
        self._siteStore = siteStore


    def getInstalledOfferingNames(self):
        """
        Get the I{offeringName} attribute of each L{InstalledOffering} in
        C{self._siteStore}.
        """
        return list(
            self._siteStore.query(InstalledOffering).getColumn("offeringName"))


    def getInstalledOfferings(self):
        """
        Return a mapping from the name of each L{InstalledOffering} in
        C{self._siteStore} to the corresponding L{IOffering} plugins.
        """
        d = {}
        installed = self._siteStore.query(InstalledOffering)
        for installation in installed:
            offering = installation.getOffering()
            if offering is not None:
                d[offering.name] = offering
        return d


    def installOffering(self, offering):
        """
        Install the given offering::

          - Create and install the powerups in its I{siteRequirements} list.
          - Create an application L{Store} and a L{LoginAccount} referring to
            it.  Install the I{appPowerups} on the application store.
          - Create an L{InstalledOffering.

        Perform all of these tasks in a transaction managed within the scope of
        this call (that means you should not call this function inside a
        transaction, or you should not handle any exceptions it raises inside
        an externally managed transaction).

        @type offering: L{IOffering}
        @param offering: The offering to install.

        @return: The C{InstalledOffering} item created.
        """
        for off in self._siteStore.query(
            InstalledOffering,
            InstalledOffering.offeringName == offering.name):
            raise OfferingAlreadyInstalled(off)

        def siteSetup():
            for (requiredInterface, requiredPowerup) in offering.siteRequirements:
                if requiredInterface is not None:
                    nn = requiredInterface(self._siteStore, None)
                    if nn is not None:
                        continue
                if requiredPowerup is None:
                    raise NotImplementedError(
                        'Interface %r required by %r but not provided by %r' %
                        (requiredInterface, offering, self._siteStore))
                self._siteStore.findOrCreate(
                    requiredPowerup, lambda p: installOn(p, self._siteStore))

            ls = self._siteStore.findOrCreate(userbase.LoginSystem)
            substoreItem = substore.SubStore.createNew(
                self._siteStore, ('app', offering.name + '.axiom'))
            ls.addAccount(offering.name, None, None, internal=True,
                          avatars=substoreItem)

            from xmantissa.publicweb import PublicWeb
            PublicWeb(store=self._siteStore, application=substoreItem,
                      prefixURL=offering.name)
            ss = substoreItem.open()
            def appSetup():
                for pup in offering.appPowerups:
                    installOn(pup(store=ss), ss)

            ss.transact(appSetup)
            # Woops, we need atomic cross-store transactions.
            io = InstalledOffering(
                store=self._siteStore, offeringName=offering.name,
                application=substoreItem)

            #Some new themes may be available now. Clear the theme cache
            #so they can show up.
            #XXX This is pretty terrible -- there
            #really should be a scheme by which ThemeCache instances can
            #be non-global. Fix this at the earliest opportunity.
            from xmantissa import webtheme
            webtheme.theThemeCache.emptyCache()
            return io
        return self._siteStore.transact(siteSetup)

registerAdapter(OfferingAdapter, Store, ixmantissa.IOfferingTechnician)



def isAppStore(s):
    """
    Return whether the given store is an application store or not.
    @param s: A Store.
    """
    if s.parent is None:
        return False
    substore = s.parent.getItemByID(s.idInParent)
    return s.parent.query(InstalledOffering,
                          InstalledOffering.application == substore
                          ).count() > 0



def getInstalledOfferingNames(s):
    """
    Return a list of the names of the Offerings which are installed on the
    given store.

    @param s: Site Store on which offering installations are tracked.
    """
    return ixmantissa.IOfferingTechnician(s).getInstalledOfferingNames()


def getInstalledOfferings(s):
    """
    Return a mapping from the names of installed IOffering plugins to
    the plugins themselves.

    @param s: Site Store on which offering installations are tracked.
    """
    return ixmantissa.IOfferingTechnician(s).getInstalledOfferings()


def installOffering(s, offering, configuration):
    """
    Create an app store for an L{Offering}, possibly installing some powerups
    on it, after checking that the site store has the requisite powerups
    installed on it. Also create an L{InstalledOffering} item referring to the
    app store and return it.
    """
    return ixmantissa.IOfferingTechnician(s).installOffering(offering)


class OfferingConfiguration(item.Item):
    """
    Provide administrative configuration tools for the L{IOffering}s available
    in this Mantissa server.
    """
    typeName = 'mantissa_offering_configuration_powerup'
    schemaVersion = 1

    installedOfferingCount = attributes.integer(default=0)
    installedOn = attributes.reference()
    powerupInterfaces = (ixmantissa.INavigableElement,)

    def installOffering(self, offering, configuration):
        """
        Create an app store for an L{Offering} and install its
        dependencies. Also create an L{InstalledOffering} in the site store,
        and return it.
        """
        s = self.store.parent
        self.installedOfferingCount += 1
        return installOffering(s, offering, configuration)


    def getTabs(self):
        # XXX profanity
        from xmantissa import webnav
        return [webnav.Tab('Admin', self.storeID, 0.3,
                           [webnav.Tab('Offerings', self.storeID, 1.0)],
                           authoritative=True)]



class UninstalledOfferingFragment(athena.LiveFragment):
    """
    Fragment representing a single Offering which has not been
    installed on the system.  It has a single remote method which will
    install it.
    """
    jsClass = u'Mantissa.Offering.UninstalledOffering'

    def __init__(self, original, offeringConfig, offeringPlugin, **kw):
        super(UninstalledOfferingFragment, self).__init__(original, **kw)
        self.offeringConfig = offeringConfig
        self.offeringPlugin = offeringPlugin

    def install(self, configuration):
        self.offeringConfig.installOffering(self.offeringPlugin, configuration)
    expose(install)


class OfferingConfigurationFragment(athena.LiveFragment):
    fragmentName = 'offering-configuration'
    live = 'athena'


    def __init__(self, *a, **kw):
        super(OfferingConfigurationFragment, self).__init__(*a, **kw)
        self.installedOfferings = getInstalledOfferingNames(self.original.store.parent)
        self.offeringPlugins = dict((p.name, p) for p in plugin.getPlugins(ixmantissa.IOffering, plugins))

    def head(self):
        return None

    def render_offerings(self, ctx, data):
        iq = inevow.IQ(ctx.tag)
        uninstalled = iq.patternGenerator('uninstalled')
        installed = iq.patternGenerator('installed')

        def offerings():
            for p in self.offeringPlugins.itervalues():
                data = {'name': p.name, 'description': p.description}
                if p.name not in self.installedOfferings:
                    f = UninstalledOfferingFragment(data, self.original, p, docFactory=loaders.stan(uninstalled()))
                    f.page = self.page
                else:
                    f = rend.Fragment(data, docFactory=loaders.stan(installed()))
                yield f

        return ctx.tag[offerings()]

registerAdapter(OfferingConfigurationFragment, OfferingConfiguration, ixmantissa.INavigableFragment)



# -*- test-case-name: xmantissa.test.test_tdb -*-

"""
This module is a deprecated version of the code present in
L{xmantissa.scrolltable}.  Look there instead.
"""

import operator
import math
import warnings

from xmantissa.ixmantissa import IColumn
from xmantissa.error import Unsortable
from xmantissa.scrolltable import AttributeColumn as _STAttributeColumn

class AttributeColumn(_STAttributeColumn):
    """
    This a deprecated name for L{xmantissa.scrolltable.AttributeColumn}.  Use
    that instead.
    """

    def __init__(self, *a, **k):
        super(AttributeColumn, self).__init__(*a, **k)
        warnings.warn("tdb.AttributeColumn is deprecated.  "
                      "Use scrolltable.AttributeColumn instead.",
                      DeprecationWarning,
                      stacklevel=2)


from axiom.attributes import AND
from axiom.queryutil import AttributeTuple

class TabularDataModel:
    """
    I represent a window onto a query that can be paged backwards and forward,
    and re-sorted.

    @ivar pageNumber: the number of the current page.

    @ivar totalPages: the total number of pages accessible to the query this
    table is browsing.

    @ivar firstItem: the first visible item

    @ivar lastItem: the last visible item

    @ivar totalItems: the total number of items accessible to the query
    this table is browsing.

    @ivar absolute: whether this pagination result is definitely correct.  in
    some cases it may be necessary to approximate or estimate the total number
    of results, and the UI should reflect this.  True if the number is
    definitely correct, False otherwise.
    """
    def __init__(self, store, primaryTableClass,
                 columns,
                 baseComparison=None,
                 itemsPerPage=20,
                 defaultSortColumn=None,
                 defaultSortAscending=True):
        """@param columns: sequence of objects adaptable to
        L{xmantissa.ixmantissa.IColumn}"""

        self.store = store
        self._currentResults = []
        self.primaryTableClass = primaryTableClass
        assert columns, "You've got to pass some columns"
        cols = self.columns = {}
        for col in map(IColumn, columns):
            if defaultSortColumn is None:
                defaultSortColumn = col.attributeID
            cols[col.attributeID] = col
        self.itemsPerPage = itemsPerPage
        self.baseComparison = baseComparison
        self.isAscending = self.defaultSortAscending = defaultSortAscending
        self.resort(defaultSortColumn,
                    defaultSortAscending)

    currentSortColumn = None # this is set in __init__ by resort()
                             # so client code will never see this value

    def resort(self, attributeID, isAscending=None):
        """Sort by one of my specified columns, identified by attributeID
        """
        if isAscending is None:
            isAscending = self.defaultSortAscending

        newSortColumn = self.columns[attributeID]
        if newSortColumn.sortAttribute() is None:
            raise Unsortable('column %r has no sort attribute' % (attributeID,))
        if self.currentSortColumn == newSortColumn:
            # if this query is to be re-sorted on the same column, but in the
            # opposite direction to our last query, then use the first item in
            # the result set as the marker
            if self.isAscending == isAscending:
                offset = 0
            else:
                # otherwise use the last
                offset = -1
        else:
            offset = 0
            self.currentSortColumn = newSortColumn
        self.isAscending = isAscending
        self._updateResults(self._sortAttributeValue(offset), True)

    def currentPage(self):
        """
        Return a sequence of mappings of attribute IDs to column values, to
        display to the user.

        nextPage/prevPage will strive never to skip items whose column values
        have not been returned by this method.

        This is best explained by a demonstration.  Let's say you have a table
        viewing an item with attributes 'a' and 'b', like this:

        oid | a | b
        ----+---+--
        0   | 1 | 2
        1   | 3 | 4
        2   | 5 | 6
        3   | 7 | 8
        4   | 9 | 0

        The table has 2 items per page.  You call currentPage and receive a
        page which contains items oid 0 and oid 1.  item oid 1 is deleted.

        If the next thing you do is to call nextPage, the result of currentPage
        following that will be items beginning with item oid 2.  This is
        because although there are no longer enough items to populate a full
        page from 0-1, the user has never seen item #2 on a page, so the 'next'
        page from the user's point of view contains #2.

        If instead, at that same point, the next thing you did was to call
        currentPage, *then* nextPage and currentPage again, the first
        currentPage results would contain items #0 and #2; the following
        currentPage results would contain items #3 and #4.  In this case, the
        user *has* seen #2 already, so the user expects to see the following
        item, not the same item again.
        """

        self._updateResults(self._sortAttributeValue(0), equalToStart=True, refresh=True)
        return self._currentResults

    def _updateResults(self, primaryColumnStart=None, equalToStart=False,
                       backwards=False, refresh=False):
        results = self._performQuery(primaryColumnStart,
                                     equalToStart,
                                     backwards)
        if not refresh and len(results) == 0:
            # If we're at the end and going forwards, or at the beginning and
            # going backwards, there are no more page results.  In these cases
            # we should hang on to our previous results, because the user is
            # still looking at the same page, and will expect the next and
            # prev buttons to do the appropriate things.  Realistically
            # speaking, it is a UI bug if this case ever occurs, since the UI
            # should disable the 'next' and 'previous' buttons using the
            # hasNextPage and hasPrevPage methods.  We gracefully handle it
            # anyway simply because we expect multiple frontends for this
            # model, and multiple frontends means lots of places for bugs.
            self.totalItems = self.totalPages = self.pageNumber = 0
            return
        self._currentResults = results
        self._paginate()

    def _determineQuery(self, primaryColumnStart, equalToStart,
                        backwards, limit):
        sortAttribute = self.currentSortColumn.sortAttribute()
        switch = (self.isAscending ^ backwards)
        # if we were sorting ascendingly, or are moving backward in the result
        # set, but not both, then sort this query ascendingly
        # use storeID as a tiebreaker, as in _sortAttributeValue
        tiebreaker = sortAttribute.type.storeID
        if switch:
            sortOrder = sortAttribute.ascending, tiebreaker.ascending
        else: # otherwise not
            sortOrder = sortAttribute.descending, tiebreaker.descending

        # if we were passed a value to use as a marker indicating our position
        # in the result set (typically it would be either the first or last
        # item from the last query)
        if primaryColumnStart is not None:
            # if we want to set the threshold at values equal to the marker
            # value we were given
            if equalToStart:
                # then use <= and >=
                ltgt = operator.__le__, operator.__ge__
            else:
                # otherwise < and >
                ltgt = operator.__lt__, operator.__gt__
            # produce an axiom comparison object by comparing the sort
            # attribute and the marker value using the appropriate operator
            offsetComparison = ltgt[switch](
                AttributeTuple(sortAttribute, tiebreaker), primaryColumnStart)
            # if we were instantiated with an additional comparison, then add
            # that to the comparison object also
            if self.baseComparison is not None:
                comparisonObj = AND(self.baseComparison,
                                    offsetComparison)
            else:
                comparisonObj = offsetComparison
        else:
            # If we've never loaded a page before, start at the beginning
            comparisonObj = self.baseComparison
        return self.store.query(self.primaryTableClass,
                                comparisonObj,
                                sort=sortOrder,
                                limit=limit)

    def _performQuery(self, primaryColumnStart=None, equalToStart=False,
                      backwards=False):
        q = self._determineQuery(primaryColumnStart, equalToStart,
                                 backwards, self.itemsPerPage)
        results = []

        for eachItem in q:
            rowDict = dict(__item__=eachItem)
            for c in self.columns.itervalues():
                rowDict[c.attributeID] = c.extractValue(self, eachItem)
            results.append(rowDict)

        if backwards:
            results.reverse()
        return results

    def _sortAttributeValue(self, offset):
        """
        return the value of the sort attribute for the item at
        'offset' in the results of the last query, otherwise None.
        """
        if self._currentResults:
            pageStart = (self._currentResults[offset][
                self.currentSortColumn.attributeID],
                         self._currentResults[offset][
                    '__item__'].storeID)
        else:
            pageStart = None
        return pageStart

    def nextPage(self):
        self._updateResults(self._sortAttributeValue(-1))

    def hasNextPage(self):
        return bool(self._performQuery(self._sortAttributeValue(-1)))

    def firstPage(self):
        self._updateResults()

    def prevPage(self):
        self._updateResults(self._sortAttributeValue(0), backwards=True)

    def hasPrevPage(self):
        return bool(self._performQuery(self._sortAttributeValue(0), backwards=True))

    def lastPage(self):
        self._updateResults(backwards=True)

    def _paginate(self):
        rslts = self._currentResults
        self.totalItems = self.store.query(self.primaryTableClass,
                                           self.baseComparison).count()
        self.totalPages = int(math.ceil(float(self.totalItems) /
                                        self.itemsPerPage))
        itemsBeforeThisPage = self._determineQuery(
            self._sortAttributeValue(0),
            equalToStart=False, backwards=True, limit=None).count()
        itemsAfterThisPage = self.totalItems - len(rslts) - itemsBeforeThisPage

        self.firstItem = itemsBeforeThisPage + 1
        self.lastItem = (self.firstItem + len(rslts)) - 1

        self.pageNumber = self.totalPages - (
            int(math.ceil(float(itemsAfterThisPage) / self.itemsPerPage)))


class PaginationParameters:

    absolute = True

    def __init__(self, tdm):
        self.tdm = tdm
        self._calculate()

    def _calculate(self):
        # XXX TODO: optimize the crap out of this by pushing its implementation
        # into nextPage/prevPage etc.
        import pprint
        pprint.pprint(self.__dict__)




# -*- test-case-name: xmantissa.test.test_theme -*-

import sys, weakref

from zope.interface import implements

from twisted.python import reflect
from twisted.python.util import sibpath
from twisted.python.filepath import FilePath
from twisted.python.components import registerAdapter

from epsilon.structlike import record

from nevow.loaders import xmlfile
from nevow import inevow, tags, athena, page

from axiom.store import Store

from xmantissa import ixmantissa
from xmantissa.ixmantissa import ITemplateNameResolver
from xmantissa.offering import getInstalledOfferings, getOfferings

class ThemeCache(object):
    """
    Collects theme information from the filesystem and caches it.

    @ivar _getAllThemesCache: a list of all themes available on the
    filesystem, or None.

    @ivar _getInstalledThemesCache: a weak-key dictionary of site
    stores to lists of themes from all installed offerings on them.
    """
    def __init__(self):
        self.emptyCache()

    def emptyCache(self):
        """
        Remove cached themes.
        """
        self._getAllThemesCache = None
        self._getInstalledThemesCache = weakref.WeakKeyDictionary()


    def _realGetAllThemes(self):
        """
        Collect themes from all available offerings.
        """
        l = []
        for offering in getOfferings():
            l.extend(offering.themes)
        l.sort(key=lambda o: o.priority)
        l.reverse()
        return l


    def getAllThemes(self):
        """
        Collect themes from all available offerings, or (if called
        multiple times) return the previously collected list.
        """
        if self._getAllThemesCache is None:
            self._getAllThemesCache = self._realGetAllThemes()
        return self._getAllThemesCache


    def _realGetInstalledThemes(self, store):
        """
        Collect themes from all offerings installed on this store.
        """
        l = []
        for offering in getInstalledOfferings(store).itervalues():
            l.extend(offering.themes)
        l.sort(key=lambda o: o.priority)
        l.reverse()
        return l


    def getInstalledThemes(self, store):
        """
        Collect themes from all offerings installed on this store, or (if called
        multiple times) return the previously collected list.
        """
        if not store in self._getInstalledThemesCache:
            self._getInstalledThemesCache[store] = (self.
                                                 _realGetInstalledThemes(store))
        return self._getInstalledThemesCache[store]


#XXX this should be local to something, not process-global.
theThemeCache = ThemeCache()
getAllThemes = theThemeCache.getAllThemes
getInstalledThemes = theThemeCache.getInstalledThemes


class SiteTemplateResolver(object):
    """
    L{SiteTemplateResolver} implements L{ITemplateNameResolver} according to
    that site's policy for loading themes.  Use this class if you are
    displaying a public page at the top level of a site.

    Currently the only available policy is to load all installed themes in
    priority order.  However, in the future, this class may provide more
    sophisticated ways of loading the preferred theme, including interacting
    with administrative tools for interactively ordering theme preference.

    @ivar siteStore: a site store (preferably one with some offerings
    installed, if you want to actually get a template back)
    """
    implements(ITemplateNameResolver)

    def __init__(self, siteStore):
        self.siteStore = siteStore


    def getDocFactory(self, name, default=None):
        """
        Locate a L{nevow.inevow.IDocFactory} object with the given name from
        the themes installed on the site store and return it.
        """
        loader = None
        for theme in getInstalledThemes(self.siteStore):
            loader = theme.getDocFactory(name)
            if loader is not None:
                return loader
        return default

registerAdapter(SiteTemplateResolver, Store, ITemplateNameResolver)


_marker = object()
def _realGetLoader(n, default=_marker):
    """
    Search all themes for a template named C{n}, returning a loader
    for it if found. If not found and a default is passed, the default
    will be returned. Otherwise C{RuntimeError} will be raised.

    This function is deprecated in favor of using a L{ThemedElement}
    for your view code, or calling
    ITemplateNameResolver(userStore).getDocFactory.
    """
    for t in getAllThemes():
        fact = t.getDocFactory(n, None)
        if fact is not None:
            return fact
    if default is _marker:
        raise RuntimeError("No loader for %r anywhere" % (n,))
    return default

_loaderCache = {}
def _memoizedGetLoader(n, default=_marker):
    """

    Find a loader for a template named C{n}, returning C{default} if
    it is provided. Otherwise raise an error. This function caches
    loaders it finds.

    This function is deprecated in favor of using a L{ThemedElement}
    for your view code, or calling
    ITemplateNameResolver(userStore).getDocFactory.
    """
    if n not in _loaderCache:
        _loaderCache[n] = _realGetLoader(n, default)
    return _loaderCache[n]

getLoader = _memoizedGetLoader

class XHTMLDirectoryTheme(object):
    """
    I am a theme made up of a directory full of XHTML templates.

    The suggested use for this class is to make a subclass,
    C{YourThemeSubclass}, in a module in your Mantissa package, create a
    directory in your package called 'yourpackage/themes/<your theme name>',
    and then pass <your theme name> as the constructor to C{YourThemeSubclass}
    when passing it to the constructor of L{xmantissa.offering.Offering}.  You
    can then avoid calculating the path name in the constructor, since it will
    be calculated based on where your subclass was defined.

    @ivar directoryName: the name of the directory containing the appropriate
        template files.

    @ivar themeName: the name of the theme that this directory represents.
        This will be displayed to the user.

    @ivar stylesheetLocation: A C{list} of C{str} giving the path segments
        beneath the root at which the stylesheet for this theme is available,
        or C{None} if there is no applicable stylesheet.
    """
    implements(ixmantissa.ITemplateNameResolver)

    stylesheetLocation = None

    def __init__(self, themeName, priority=0, directoryName=None):
        """
        Create a theme based off of a directory full of XHTML templates.

        @param themeName: sets my themeName

        @param priority: an integer that affects the ordering of themes
        returned from L{getAllThemes}.

        @param directoryName: If None, calculates the directory name based on
        the module the class is defined in and the given theme name.  For a
        subclass C{bar.baz.FooTheme} defined in C{bar/baz.py} the instance
        C{FooTheme('qux')}, regardless of where it is created, will have a
        default directoryName of {bar/themes/qux/}.
        """

        self.themeName = themeName
        self.priority = priority
        self.cachedLoaders = {}
        if directoryName is None:
            self.directory = FilePath(
                sibpath(sys.modules[self.__class__.__module__].__file__,
                        'themes') + '/' + self.themeName)
        else:
            self.directory = FilePath(directoryName)
        self.directoryName = self.directory.path


    def head(self, request, website):
        """
        Provide content to include in the head of the document.  If you only
        need to provide a stylesheet, see L{stylesheetLocation}.  Otherwise,
        override this.

        @type request: L{inevow.IRequest} provider
        @param request: The request object for which this is a response.

        @param website: The site-wide L{xmantissa.website.WebSite} instance.
            Primarily of interest for its C{rootURL} method.

        @return: Anything providing or adaptable to L{nevow.inevow.IRenderer},
            or C{None} to include nothing.
        """
        stylesheet = self.stylesheetLocation
        if stylesheet is not None:
            root = website.rootURL(request)
            for segment in stylesheet:
                root = root.child(segment)
            return tags.link(rel='stylesheet', type='text/css', href=root)


    # ITemplateNameResolver
    def getDocFactory(self, fragmentName, default=None):
        """
        For a given fragment, return a loaded Nevow template.

        @param fragmentName: the name of the template (can include relative
        paths).

        @param default: a default loader; only used if provided and the
        given fragment name cannot be resolved.

        @return: A loaded Nevow template.
        @type return: L{nevow.loaders.xmlfile}
        """
        if fragmentName in self.cachedLoaders:
            return self.cachedLoaders[fragmentName]
        segments = fragmentName.split('/')
        segments[-1] += '.html'
        file = self.directory
        for segment in segments:
            file = file.child(segment)
        if file.exists():
            loader = xmlfile(file.path)
            self.cachedLoaders[fragmentName] = loader
            return loader
        return default



class MantissaTheme(XHTMLDirectoryTheme):
    """
    Basic Mantissa-provided theme.

    Most templates in the I{base} theme will usually be satisfied from this
    theme provider.
    """
    stylesheetLocation = ['static', 'mantissa-base', 'mantissa.css']



class ThemedDocumentFactory(record('fragmentName resolverAttribute')):
    """
    A descriptor which finds a template loader based on the Mantissa theme
    system.

    Use this to have the appropriate template loaded automatically at the time
    of first access::

        class YourElement(Element):
            '''
            An element for showing something.
            '''
            docFactory = ThemedDocumentFactory('template-name', 'userStore')

            def __init__(self, store):
                self.userStore = store


    When C{docFactory} is looked up on an instance of YourElement, the
    ITemplateNameResolver powerup on its C{'userStore'} attribute will
    be loaded and asked for the C{'template-name'} template loader.  The
    C{docFactory} attribute access will result in the object in which
    this results.

    @ivar fragmentName: A C{str} giving the name of the template which should
        be loaded.
    @ivar resolverAttribute: A C{str} giving the name of the attribute on
        instances which is adaptable to L{ITemplateNameResolver}.
    """
    def __get__(self, oself, type):
        """
        Get the document loader for C{self.fragmentName}.
        """
        resolver = ITemplateNameResolver(
            getattr(oself, self.resolverAttribute))
        return resolver.getDocFactory(self.fragmentName)



class _ThemedMixin(object):
    """
    Mixin for L{nevow.inevow.IRenderer} implementations which want to use the
    theme system.
    """

    implements(ixmantissa.ITemplateNameResolver)

    def __init__(self, fragmentParent=None):
        """
        Create a themed fragment with the given parent.

        @param fragmentParent: An object to pass to C{setFragmentParent}.  If
        not None, C{self.setFragmentParent} is called immediately.  It is
        suggested but not required that you set this here; if not, the
        resulting fragment will be initialized in an inconsistent state.  You
        must call setFragmentParent to correct this before this fragment is
        rendered.
        """
        super(_ThemedMixin, self).__init__()
        if fragmentParent is not None:
            self.setFragmentParent(fragmentParent)


    def head(self):
        """
        Don't do anything.
        """

    def rend(self, context, data):
        """
        Automatically retrieve my C{docFactory} based on C{self.fragmentName}
        before invoking L{athena.LiveElement.rend}.
        """
        if self.docFactory is None:
            self.docFactory = self.getDocFactory(self.fragmentName)
        return super(_ThemedMixin, self).rend(context, data)


    def pythonClass(self, request, tag):
        """
        This renderer is available on all themed fragments.  It returns the fully
        qualified python name of the class of the fragment being rendered.
        """
        return reflect.qual(self.__class__)
    page.renderer(pythonClass)


    def render_pythonClass(self, ctx, data):
        return self.pythonClass(inevow.IRequest(ctx), ctx.tag)


    # ITemplateNameResolver
    def getDocFactory(self, fragmentName, default=None):
        f = getattr(self.page, "getDocFactory", getLoader)
        return f(fragmentName, default)



class ThemedFragment(_ThemedMixin, athena.LiveFragment):
    """
    Subclass me to create a LiveFragment which supports automatic
    theming. (Deprecated)

    @ivar fragmentName: A short string naming the template from which the
    docFactory for this fragment should be loaded.

    @see ThemedElement
    """
    fragmentName = 'fragment-no-fragment-name-specified'



class ThemedElement(_ThemedMixin, athena.LiveElement):
    """
    Subclass me to create a LiveElement which supports automatic theming.

    @ivar fragmentName: A short string naming the template from which the
    docFactory for this fragment should be loaded.
    """
    fragmentName = 'element-no-fragment-name-specified'



# -*- test-case-name: xmantissa.test.test_q2q -*-

from axiom import item, attributes

from vertex import q2qclient

from xmantissa import ixmantissa

class UniversalEndpointService(item.Item):
    """
    """
    typeName = 'mantissa_q2q_service'
    schemaVersion = 1

    q2qPortNumber = attributes.integer(default=8787)
    inboundTCPPortNumber = attributes.integer(default=None)
    publicIP = attributes.integer(default=None)
    udpEnabled = attributes.integer(default=False)
    certificatePath = attributes.path(default=None)

    _svcInst = attributes.inmemory()

    def __init__(self, **kw):
        super(UniversalEndpointService, self).__init__(**kw)
        if self.certificatePath is None:
            self.certificatePath = self.store.newDirectory(str(self.storeID), 'certificates')
        if not self.certificatePath.exists():
            self.certificatePath.makedirs()


    def activate(self):
        self._svcInst = None

    def _makeService(self):
        self._svcInst = q2qclient.ClientQ2QService(
            self.certificatePath.path,
            publicIP=self.publicIP,
            inboundTCPPortnum=self.inboundTCPPortNumber,
            udpEnabled=self.udpEnabled,
            )

    def _getService(self):
        if self._svcInst is None:
            self._makeService()
        return self._svcInst

    def installOn(self, other):
        other.powerUp(self, ixmantissa.IQ2QService)

    def listenQ2Q(self, fromAddress, protocolsToFactories, serverDescription):
        return self._getService().listenQ2Q(fromAddress,
                                       protocolsToFactories,
                                       serverDescription)

    def connectQ2Q(self, fromAddress, toAddress, protocolName,
                   protocolFactory, usePrivateCertificate=None,
                   fakeFromDomain=None, chooser=None):
        return self._getService().connectQ2Q(
            fromAddress, toAddress, protocolName, protocolFactory,
            usePrivateCertificate, fakeFromDomain, chooser)



# -*- test-case-name: xmantissa.test.historic.test_website3to4 -*-

from axiom.userbase import LoginSystem
from axiom.test.historic.stubloader import saveStub

from axiom.plugins.mantissacmd import Mantissa

from xmantissa.website import WebSite

cert = (
    '-----BEGIN CERTIFICATE-----\n'
    'MIICmjCCAgMCBACFAjkwDQYJKoZIhvcNAQEEBQAwgZMxCzAJBgNVBAYTAlVTMRMw\n'
    'EQYDVQQDEwpkaXZtb2QuY29tMREwDwYDVQQHEwhOZXcgWW9yazETMBEGA1UEChMK\n'
    'RGl2bW9kIExMQzERMA8GA1UECBMITmV3IFlvcmsxITAfBgkqhkiG9w0BCQEWEnN1\n'
    'cHBvcnRAZGl2bW9kLm9yZzERMA8GA1UECxMIU2VjdXJpdHkwHhcNMDgwMjIwMjEy\n'
    'NDExWhcNMDkwMjE5MjEyNDExWjCBkzELMAkGA1UEBhMCVVMxEzARBgNVBAMTCmRp\n'
    'dm1vZC5jb20xETAPBgNVBAcTCE5ldyBZb3JrMRMwEQYDVQQKEwpEaXZtb2QgTExD\n'
    'MREwDwYDVQQIEwhOZXcgWW9yazEhMB8GCSqGSIb3DQEJARYSc3VwcG9ydEBkaXZt\n'
    'b2Qub3JnMREwDwYDVQQLEwhTZWN1cml0eTCBnzANBgkqhkiG9w0BAQEFAAOBjQAw\n'
    'gYkCgYEA3ucaT8gUB7BEp2dfRulWBRT6tTDELA7sJzyk+12E1vxQppJDzwG8VgSj\n'
    'sOl8Jw0qnUb/Qoe96UlA8hDYBbmwz0CCvVRSj+1GYj6Ka8NeheME2RU3/benLbyL\n'
    'S7HUQ93Mqs3VWrlv2lMbgp29njwJqvqMRt8JGB1ql8xUDSLw4kcCAwEAATANBgkq\n'
    'hkiG9w0BAQQFAAOBgQAXxMBJu+VkazQSuOnIn5Ewug2tHmf0sxT7FkcB2nEviQ7U\n'
    'e2bb95IL9XqkO0yKEbJ5K8T8SyXW9VNUATce4JO6NNikyVCZzV1dG2+ATDBaaVHK\n'
    'S2Beh1p6boFvv0+k2qZ/9JmJYVx4l1xPavc70x95rR2E0kuwhyw4miHpSMqfpA==\n'
    '-----END CERTIFICATE-----\n'
    '-----BEGIN RSA PRIVATE KEY-----\n'
    'MIICXQIBAAKBgQDe5xpPyBQHsESnZ19G6VYFFPq1MMQsDuwnPKT7XYTW/FCmkkPP\n'
    'AbxWBKOw6XwnDSqdRv9Ch73pSUDyENgFubDPQIK9VFKP7UZiPoprw16F4wTZFTf9\n'
    't6ctvItLsdRD3cyqzdVauW/aUxuCnb2ePAmq+oxG3wkYHWqXzFQNIvDiRwIDAQAB\n'
    'AoGAG/YHgeyKPrCo3AsGk6GfjcGk9WeppBE3JHDiDToc+M7r2wlMAkKoem3Yjs+r\n'
    'KEbpipMmYBUhCIuM3xCn2IgDmq/9rC+mDmEu7mEvL0Rnl5Ns6m/uw61kYKDAghYg\n'
    'K7lD3jlAT/a9I8wB2UO9F6p8166YERU736Qa4GUle4l8irECQQD0V6ZbbW1o5j5s\n'
    'IUzhVvBr/flWabpMJ9Vw3eLy695iFjgx+5W0nD+JK1ny8MiwCRsjoRTXldHhdaod\n'
    '8VbPz/QJAkEA6YmX6XksIb8JUYFtPk0WodQmz51qzo0jol3COL/rXuPVkTcesyTM\n'
    '61S7WSv0G6pMqE9xw0llMBON7Pr24N/XzwJBALW+eFvrEgWDtQyi3FeEXkJFX+/5\n'
    'pnu86VMRiByeewREeLoc4ya7TbsOxtIgbXYa39fpmeIda0ajSc0J1UOv71kCQQCO\n'
    'q20vx8PrNc7WiTAY4HVUFcxEB5Ipb1X2qjqt+qkrBhsBpN/PZ0r89X2iw1RU1lwQ\n'
    'csA4Io17qmaJAORziqxHAkAb2zin9SzS58+X55pGVp8PwhGLmm9cGH/DtWVSIAl2\n'
    'q3pqCmcxnimc+IYJJlY6dkk7jtnIVTWz3B9XUOtKGEYF\n'
    '-----END RSA PRIVATE KEY-----\n')

def createDatabase(store):
    """
    Initialize the given Store for use as a Mantissa webserver.
    """
    Mantissa().installSite(store, u'')
    ws = store.findUnique(WebSite)
    ws.portNumber = 8088
    ws.securePortNumber = 6443
    ws.certificateFile = 'path/to/cert.pem'
    certPath = store.dbdir.child('path').child('to').child('cert.pem')
    certPath.parent().makedirs()
    fObj = certPath.open('w')
    fObj.write(cert)
    fObj.close()
    ws.httpLog = 'path/to/httpd.log'
    ws.hitCount = 123

    loginSystem = store.findUnique(LoginSystem)
    account = loginSystem.addAccount(u'testuser', u'localhost', None)
    subStore = account.avatars.open()
    WebSite(store=subStore, hitCount=321).installOn(subStore)


if __name__ == '__main__':
    saveStub(createDatabase, 7617)
